
/*
 *    miracleshaper - Traffic shaper and prioritizer for DSLs.
 *    Copyright (C) 2007  Florian Zumbiehl <florz@florz.de>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

#include <stdlib.h>
#include <sys/poll.h>
#include <stdio.h>
#include <errno.h>

#include "utils.h"
#include "event.h"
#include "config.h"

struct event_readcallback {
	void (*callback)(int,void *);
	void *arg;
};

static struct pollfd *event_pollfds;
static struct event_readcallback *event_readcallbacks;
schedtime_t event_timer_hz;
struct event_timer *event_timers;
static int event_numfds,event_numtimers;

void event_init(){
	FILE *f;
	char buf[128];
	unsigned x,y;

	if(!(f=fopen("/proc/cpuinfo","r")))die(true,"fopen(\"/proc/cpuinfo\")");
	while(fgets(buf,sizeof(buf),f)){
		if(sscanf(buf,"cpu MHz : %u.%u",&x,&y)==2){
			if(fclose(f))die(true,"fclose(\"/proc/cpuinfo\")");
			event_timer_hz=(schedtime_t)x*1000000+y*1000;
			return;
		}
	}
	die(false,"Couldn't find CPU clock in /proc/cpuinfo");
}

void event_add_readfd(int fd,void (*callback)(int,void *),void *arg){
	void *new;

	new=realloc(event_pollfds,sizeof(*event_pollfds)*(event_numfds+1));
	if(!new)die(false,"realloc(event_pollfds) failed");
	event_pollfds=new;
	event_pollfds[event_numfds].fd=fd;
	event_pollfds[event_numfds].events=POLLIN;
	new=realloc(event_readcallbacks,sizeof(*event_readcallbacks)*(event_numfds+1));
	if(!new)die(false,"realloc(event_readcallbacks) failed");
	event_readcallbacks=new;
	event_readcallbacks[event_numfds].callback=callback;
	event_readcallbacks[event_numfds].arg=arg;
	event_numfds++;
}

void event_timer_set_delay(int timer,schedtime_t time){
	event_timers[timer].time=event_timer_now()+time;
}

int event_timer_create(void (*callback)(void *),void *arg){
	void *new;

	new=realloc(event_timers,sizeof(*event_timers)*(event_numtimers+1));
	if(!new)die(false,"realloc(event_timers) failed");
	event_timers=new;
	event_timers[event_numtimers].active=false;
	event_timers[event_numtimers].time=0;
	event_timers[event_numtimers].callback=callback;
	event_timers[event_numtimers].arg=arg;
	return event_numtimers++;
}

void event_dispatch(){
	while(true){
		schedtime_t to;

		to=-1;
		for(int x=event_numtimers;x--;)
			if(event_timers[x].active&&(to<0||event_timers[x].time<to))
				to=event_timers[x].time;
		if(to>0){
			to-=event_timer_now()+SCHEDULING_JITTER;
			if(to<=0)to=0;
				else to=to*1000/event_timer_hz+1+1; // add one for rounding up and one to compensate for poll() sleeping too short
		}
		if(to)
			if(poll(event_pollfds,event_numfds,to)<0)
				if(errno!=EINTR)die(true,"poll()");
		for(int x=event_numtimers;x--;)
			while(event_timers[x].active&&event_timers[x].time<event_timer_now()+SCHEDULING_JITTER)
				event_timers[x].callback(event_timers[x].arg);
		if(to)
			for(int x=event_numfds;x--;)
				if(event_pollfds[x].revents&(POLLIN|POLLERR))
					event_readcallbacks[x].callback(event_pollfds[x].fd,event_readcallbacks[x].arg);
	}
}

