
/*
 *    miracleshaper - Traffic shaper and prioritizer for DSLs.
 *    Copyright (C) 2007  Florian Zumbiehl <florz@florz.de>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

#include <net/ethernet.h>

#include "pkt.h"

/*** IP ******************************************************************/

bool pkt_ip_update_checksum(struct pkt *pkt,unsigned int old,unsigned int new){
	uint16_t *chksump;
	unsigned long chksum;

	if(pkt_get_plen(pkt)<10+2)return false;
	chksump=(uint16_t *)(pkt_get_pbuf(pkt)+10);
	chksum=*chksump;
	chksum=(chksum^0xfffful)+(old^0xfffful)+new;
	while(chksum&~0xfffful)chksum=(chksum&0xfffful)+(chksum>>16);
	*chksump=chksum^0xfffful;
	return true;
}

bool pkt_ip_mark_ecn(struct pkt *pkt){
	uint8_t *tosp;

	if(pkt_get_plen(pkt)<0+2)return false;
	tosp=(uint8_t *)(pkt_get_pbuf(pkt)+1);
	if(*tosp&3){
		unsigned long oldval;

		oldval=*(uint16_t *)(pkt_get_pbuf(pkt)+0);
		*tosp|=3;
		return pkt_ip_update_checksum(pkt,oldval,*(uint16_t *)(pkt_get_pbuf(pkt)+0));
	}
	return false;
}

/*** IPv6 ****************************************************************/

bool pkt_ip6_mark_ecn(struct pkt *pkt){
	uint8_t *tcp;

	if(pkt_get_plen(pkt)<0+2)return false;
	tcp=(uint8_t *)(pkt_get_pbuf(pkt)+1);
	if(*tcp&48){
		*tcp|=48;
		return true;
	}
	return false;
}

/*** IP and IPv6 *********************************************************/

bool pkt_mark_ecn(struct pkt *pkt){
	long eth_proto;

	pkt_pinit(pkt);
	if((eth_proto=pkt_eth_get_proto(pkt))<0)goto error;
	if(eth_proto==ETH_P_PPP_SES){
		long ppp_proto;

		if(!pkt_eth_nextlayer(pkt))goto error;
		if(!pkt_pppoeses_nextlayer(pkt))goto error;
		if((ppp_proto=pkt_ppp_get_proto(pkt))<0)goto error;
		if(!pkt_ppp_nextlayer(pkt))goto error;
		if(ppp_proto==PPP_P_IP){
			if(!pkt_ip_mark_ecn(pkt))goto error;
			return true;
		}
		if(ppp_proto==PPP_P_IPV6){
			if(!pkt_ip6_mark_ecn(pkt))goto error;
			return true;
		}
	}
error:
	return false;
}

